import logging

from .metric_directionality import is_better_operator
import pandas as pd
from datetime import datetime

def _past_threshold(df, metric_filter):
    return df[is_better_operator(metric_filter['metric'])(
        df['raw_value'],
        metric_filter['threshold_value']
    )]


def _close_to_best_case(df, metric_filter):
    return df[df['dist_from_best_case'] < metric_filter['max_from_best']]


def _of_metric(df, metric_filter):
    return df[
        (df['metric'] == metric_filter['metric']) &
        (df['parameter'] == metric_filter['parameter'])
    ]


def model_groups_filter(
    train_end_times,
    initial_model_group_ids,
    models_table,
    db_engine
):
    """Before creating the distance table, we want to filter the model_group_ids
    and train_end_times only if they are reasonable -- every model group should
    have the same train_end_times -- to prevent incomparable model groups from
    being populated to the distance table.

    Args:
        train_end_times (list) The set of train_end_times to consider during
                               the iteration
        initial_model_group_ids (list) The initial list of model group ids to
                                       narrow down
        models_table (string) The name of the results schema
        db_engine (sqlalchemy.engine) A database engine with access to results
                                      schema of a completed modeling run
    """

    if isinstance(train_end_times, str) or not hasattr(train_end_times, '__iter__'):
        raise TypeError("train_end_times should be a list of str or timestamp")

    if not bool(train_end_times):
        raise ValueError("train_end_times shouldn't be an empty list")

    end_times_sql = "ARRAY[{}]".format(
        ', '.join(
            "'{}'".format(
                end_time.strftime('%Y-%m-%d') if isinstance(end_time, datetime) else end_time
            ) for end_time in train_end_times
        )
    )
    logging.info("Checking if all model groups have the same train end times")
    logging.info(f"Found {len(initial_model_group_ids)} total model groups")
    query = f"""
        SELECT model_group_id
        FROM (
            SELECT
                model_group_id,
                array_agg(distinct train_end_time::date::text) as train_end_time_list
            FROM model_metadata.{models_table}
            WHERE model_group_id in {tuple(initial_model_group_ids)}
            GROUP BY model_group_id
        ) as t
        WHERE train_end_time_list = {end_times_sql}
        """
    model_groups_df = pd.read_sql(query, con=db_engine)
    model_group_ids = list(model_groups_df['model_group_id'])

    dropped_model_groups = len(initial_model_group_ids) - len(model_group_ids)
    logging.info(f"Dropped {dropped_model_groups} model groups which don't have the same train end times")
    logging.info(f"Found {len(model_group_ids)} total model groups past the checker")

    return set(model_group_ids)

class ModelGroupThresholder(object):

    def __init__(
        self,
        distance_from_best_table,
        train_end_times,
        initial_model_group_ids,
        initial_metric_filters
    ):
        """Iteratively narrow down a list of model groups by changing thresholds
        for max below best model and minimum absolute value with respect to
        different metrics

        Args:
            distance_from_best_table (audition.DistanceFromBestTable)
                A pre-populated distance-from-best database table
            train_end_times (list) The set of train end times to consider during iteration
            initial_model_group_ids (list) The initial list of model group ids to
                narrow down

        """
        self.distance_from_best_table = distance_from_best_table
        self.train_end_times = train_end_times
        self._initial_model_group_ids = initial_model_group_ids
        self._metric_filters = initial_metric_filters


    def _filter_model_groups(self, df, filter_func):
        """Filter model groups by ensuring each of their metrics meets the given
            filtering function.

        Args:
            df (pandas.DataFrame): A set of rows in the format given by
                audition.DistanceFromBestTable.as_dataframe
            filter_func (function): A function that takes a dataframe and a
                metric filter and returns a subset of the dataframe

        Returns: (set) The model group ids that pass filtering
        """
        passing = set(self._initial_model_group_ids)
        for metric_filter in self._metric_filters:
            passing &= set(filter_func(
                _of_metric(df, metric_filter),
                metric_filter
            )['model_group_id'])
        return passing

    def model_groups_past_threshold(self, df):
        """Return the model groups in the dataframe that are above the
            currently-configured minimum value

        Args:
            df (pandas.DataFrame): A set of rows in the format given by
                audition.DistanceFromBestTable.as_dataframe
        Returns: (set) The model group ids above the minimum value for each metric
        """
        return self._filter_model_groups(df, _past_threshold)

    def model_groups_close_to_best_case(self, df):
        """Return the model groups in the dataframe that are close enough to
            the best value according to current metric filter configuration

        Args:
            df (pandas.DataFrame): A set of rows in the format given by
                audition.DistanceFromBestTable.as_dataframe
        Returns: (set) The model group ids close to the best value for each metric
        """
        return self._filter_model_groups(df, _close_to_best_case)

    def model_groups_passing_rules(self):
        """Return the model groups passing both the close-to-best and
            above-min checks based on the current filters.

        Works by ensuring that a model group passes the close-to-best check
        for at least one train end time, and that is passes the above-min
        check for *all* train end times. Model groups must pass both
        of these checks.

        Returns: (set) The passing model group ids
        """
        past_threshold_model_groups = set(self._initial_model_group_ids)
        close_to_best_model_groups = set()
        for train_end_time in self.train_end_times:
            df_as_of = self.distance_from_best_table.dataframe_as_of(
                model_group_ids=self._initial_model_group_ids,
                train_end_time=train_end_time,
            )
            close_to_best = self.model_groups_close_to_best_case(df_as_of)
            logging.info(
                'Found %s model groups close to best for %s',
                len(close_to_best),
                train_end_time
            )
            close_to_best_model_groups |= close_to_best

            past_threshold = self.model_groups_past_threshold(df_as_of)
            logging.info(
                'Found %s model groups above min for %s',
                len(past_threshold),
                train_end_time
            )
            past_threshold_model_groups &= past_threshold

        total_model_groups = close_to_best_model_groups & past_threshold_model_groups
        logging.info(
            'Found %s total model groups past threshold',
            len(total_model_groups)
        )
        return total_model_groups


    def update_filters(self, new_metric_filters):
        """Update the saved metric filters.

        Args: new_metric_filters (list) A list of metrics to filter model
            groups on, and how to filter them. Each entry should be a dict
            with the keys:

                metric (string) -- model evaluation metric, such as 'precision@'
                parameter (string) -- model evaluation metric parameter,
                    such as '300_abs'
                max_below_best (float) The maximum value that the given metric
                    can be below the best for a given train end time
                min_value (float) The minimum value that the given metric can be
        """
        if new_metric_filters != self._metric_filters:
            self._metric_filters = new_metric_filters

    @property
    def model_group_ids(self):
        return self.model_groups_passing_rules()



import datetime
import shutil
import sys
import random
from contextlib import contextmanager
import functools
import operator
import tempfile

import sqlalchemy

import pandas as pd
import yaml
import numpy
from sqlalchemy.orm import sessionmaker

from triage.component.results_schema import Model


def str_in_sql(values):
    return ','.join(map(lambda x: "'{}'".format(x), values))


def feature_list(feature_dictionary):
    """Convert a feature dictionary to a sorted list

    Args: feature_dictionary (dict)

    Returns: sorted list of feature names
    """
    if not feature_dictionary:
        return []
    return sorted(
        functools.reduce(
            operator.concat,
            (feature_dictionary[key] for key in feature_dictionary.keys())
        )
    )


def convert_string_column_to_date(column):
    return(
        [datetime.datetime.strptime(date, '%Y-%m-%d').date() for date in column]
    )


def create_schemas(engine, features_tables, labels, states):
    """ This function makes a features schema and populates it with the fake
    data from above.

    :param engine: a postgresql engine
    :type engine: sqlalchemy.engine
    """
    # create features schema and associated tables
    engine.execute('drop schema if exists features cascade; create schema features;')
    for table_number, table in enumerate(features_tables):
        create_features_table(table_number, table, engine)
    # create labels schema and table
    engine.execute('drop schema if exists labels cascade; create schema labels;')
    engine.execute(
        """
            create table labels.labels (
                entity_id int,
                as_of_date date,
                label_timespan interval,
                label_name char(30),
                label_type char(30),
                label int
            )
        """
    )
    for row in labels:
        engine.execute(
            'insert into labels.labels values (%s, %s, %s, %s, %s, %s)',
            row
        )

    # create sparse state table
    engine.execute('drop schema if exists staging cascade; create schema staging;')
    engine.execute(
        """
            create table staging.sparse_states (
                entity_id int,
                as_of_date date,
                state_one bool,
                state_two bool
            )
        """
    )
    for row in states:
        engine.execute(
            'insert into staging.sparse_states values (%s, %s, %s, %s)',
            row
        )


def create_features_table(table_number, table, engine):
    engine.execute(
        """
            create table features.features{} (
                entity_id int, as_of_date date, f{} int, f{} int
            )
        """.format(table_number, (table_number*2)+1, (table_number*2)+2)
    )
    for row in table:
        engine.execute(
            """
                insert into features.features{} values (%s, %s, %s, %s)
            """.format(table_number),
            row
        )


def create_entity_date_df(
    labels,
    states,
    as_of_dates,
    state_one,
    state_two,
    label_name,
    label_type,
    label_timespan
):
    """ This function makes a pandas DataFrame that mimics the entity-date table
    for testing against.
    """
    0, '2016-02-01', '1 month', 'booking', 'binary', 0
    labels_table = pd.DataFrame(labels, columns=[
        'entity_id',
        'as_of_date',
        'label_timespan',
        'label_name',
        'label_type',
        'label'
    ])
    states_table = pd.DataFrame(states, columns=[
        'entity_id',
        'as_of_date',
        'state_one',
        'state_two'
    ]).set_index(['entity_id', 'as_of_date'])
    as_of_dates = [date.date() for date in as_of_dates]
    labels_table = labels_table[labels_table['label_name'] == label_name]
    labels_table = labels_table[labels_table['label_type'] == label_type]
    labels_table = labels_table[labels_table['label_timespan'] == label_timespan]
    labels_table = labels_table.join(
        other=states_table,
        on=('entity_id', 'as_of_date'),
    )
    labels_table = labels_table[labels_table['state_one'] & labels_table['state_two']]
    ids_dates = labels_table[['entity_id', 'as_of_date']]
    ids_dates = ids_dates.sort_values(['entity_id', 'as_of_date'])
    ids_dates['as_of_date'] = [datetime.datetime.strptime(
        date,
        '%Y-%m-%d'
    ).date() for date in ids_dates['as_of_date']]
    ids_dates = ids_dates[ids_dates['as_of_date'].isin(as_of_dates)]
    print(ids_dates)

    return(ids_dates.reset_index(drop=True))


def NamedTempFile():
    if sys.version_info >= (3, 0, 0):
        return tempfile.NamedTemporaryFile(mode='w+', newline='')
    else:
        return tempfile.NamedTemporaryFile()


@contextmanager
def TemporaryDirectory():
    name = tempfile.mkdtemp()
    try:
        yield name
    finally:
        shutil.rmtree(name)


@contextmanager
def fake_metta(matrix_dict, metadata):
    """Stores matrix and metadata in a metta-data-like form

    Args:
    matrix_dict (dict) of form { columns: values }.
        Expects an entity_id to be present which it will use as the index
    metadata (dict). Any metadata that should be set

    Yields:
        tuple of filenames for matrix and metadata
    """
    matrix = pd.DataFrame.from_dict(matrix_dict).set_index('entity_id')
    with tempfile.NamedTemporaryFile() as matrix_file:
        with tempfile.NamedTemporaryFile('w') as metadata_file:
            hdf = pd.HDFStore(matrix_file.name)
            hdf.put('title', matrix, data_columns=True)
            matrix_file.seek(0)

            yaml.dump(metadata, metadata_file)
            metadata_file.seek(0)
            yield (matrix_file.name, metadata_file.name)


def fake_labels(length):
    return numpy.array([random.choice([True, False]) for i in range(0, length)])


class MockTrainedModel(object):
    def predict_proba(self, dataset):
        return numpy.random.rand(len(dataset), len(dataset))


def fake_trained_model(project_path, model_storage_engine, db_engine):
    """Creates and stores a trivial trained model

    Args:
        project_path (string) a desired fs/s3 project path
        model_storage_engine (triage.storage.ModelStorageEngine)
        db_engine (sqlalchemy.engine)

    Returns:
        (int) model id for database retrieval
    """
    trained_model = MockTrainedModel()
    model_storage_engine.write(trained_model, 'abcd')
    session = sessionmaker(db_engine)()
    db_model = Model(model_hash='abcd')
    session.add(db_model)
    session.commit()
    return trained_model, db_model.model_id


def assert_index(engine, table, column):
    """Assert that a table has an index on a given column

    Does not care which position the column is in the index
    Modified from https://www.gab.lc/articles/index_on_id_with_postgresql

    Args:
        engine (sqlalchemy.engine) a database engine
        table (string) the name of a table
        column (string) the name of a column
    """
    query = """
        SELECT 1
        FROM pg_class t
             JOIN pg_index ix ON t.oid = ix.indrelid
             JOIN pg_class i ON i.oid = ix.indexrelid
             JOIN pg_attribute a ON a.attrelid = t.oid
        WHERE
             a.attnum = ANY(ix.indkey) AND
             t.relkind = 'r' AND
             t.relname = '{table_name}' AND
             a.attname = '{column_name}'
    """.format(
        table_name=table,
        column_name=column
    )
    num_results = len([row for row in engine.execute(query)])
    assert num_results >= 1


def create_dense_state_table(db_engine, table_name, data):
    db_engine.execute('''create table {} (
        entity_id int,
        state text,
        start_time timestamp,
        end_time timestamp
    )'''.format(table_name))

    for row in data:
        db_engine.execute(
            'insert into {} values (%s, %s, %s, %s)'.format(table_name),
            row
        )


def create_binary_outcome_events(db_engine, table_name, events_data):
    db_engine.execute(
        'create table events (entity_id int, outcome_date date, outcome bool)'
    )
    for event in events_data:
        db_engine.execute(
            'insert into {} values (%s, %s, %s::bool)'.format(table_name),
            event
        )


def retry_if_db_error(exception):
    return isinstance(exception, sqlalchemy.exc.OperationalError)



from unittest.mock import patch

import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.regrets import SelectionRulePicker
from triage.component.audition.selection_rule_performance import SelectionRulePerformancePlotter
from triage.component.audition.selection_rules import BoundSelectionRule

from .utils import create_sample_distance_table


TRAIN_END_TIMES = ['2014-01-01', '2015-01-01']


class MockSelectionRulePicker(object):
    def results_for_rule(*args, **kwargs):
        return [
            {
                'train_end_time': TRAIN_END_TIMES[0],
                'dist_from_best_case_next_time': 0.15,
                'raw_value_next_time': 0.5,
                'model_group_id': 1
            },
            {
                'train_end_time': TRAIN_END_TIMES[1],
                'dist_from_best_case_next_time': 0.30,
                'raw_value_next_time': 0.4,
                'model_group_id': 2
            }
        ]


def test_SelectionRulePerformancePlotter_generate_plot_data():
    plotter = SelectionRulePerformancePlotter(MockSelectionRulePicker())
    df = plotter.generate_plot_data(
        bound_selection_rules=[BoundSelectionRule(
            function_name='best_current_value',
            args={'metric': 'precision@', 'parameter': '100_abs'},
        ), BoundSelectionRule(
            function_name='best_average_value',
            args={'metric': 'precision@', 'parameter': '100_abs'},
        )],
        regret_metric='precision@',
        regret_parameter='100_abs',
        model_group_ids=[1, 2],
        train_end_times=TRAIN_END_TIMES,
    )
    print(df.to_dict('list'))
    assert df.to_dict('list') == {
        'selection_rule': [
            'best_current_value_precision@_100_abs',
            'best_current_value_precision@_100_abs',
            'best_average_value_precision@_100_abs',
            'best_average_value_precision@_100_abs'
        ],
        'train_end_time': TRAIN_END_TIMES + TRAIN_END_TIMES,
        'regret': [0.15, 0.30, 0.15, 0.30],
        'raw_value_next_time': [0.5, 0.4, 0.5, 0.4],
        'model_group_id': [1, 2, 1, 2]
    }


def test_SelectionRulePerformancePlotter_plot_regrets():
    with patch('triage.component.audition.selection_rule_performance.plot_cats') as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = SelectionRulePerformancePlotter(SelectionRulePicker(distance_table))
            plotter.plot(
                bound_selection_rules=[BoundSelectionRule(
                    function_name='best_current_value',
                    args={'metric': 'precision@', 'parameter': '100_abs'},
                ), BoundSelectionRule(
                    function_name='best_average_value',
                    args={'metric': 'precision@', 'parameter': '100_abs'},
                )],
                regret_metric='precision@',
                regret_parameter='100_abs',
                model_group_ids=[1, 2],
                train_end_times=['2014-01-01', '2015-01-01'],
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert 'regret' in kwargs['frame']
        assert 'train_end_time' in kwargs['frame']
        assert kwargs['x_col'] == 'train_end_time'
        assert kwargs['y_col'] == 'regret'


def test_SelectionRulePerformancePlotter_plot_metrics():
    with patch('triage.component.audition.selection_rule_performance.plot_cats') as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = SelectionRulePerformancePlotter(SelectionRulePicker(distance_table))
            plotter.plot(
                bound_selection_rules=[BoundSelectionRule(
                    function_name='best_current_value',
                    args={'metric': 'precision@', 'parameter': '100_abs'},
                ), BoundSelectionRule(
                    function_name='best_average_value',
                    args={'metric': 'precision@', 'parameter': '100_abs'},
                )],
                regret_metric='precision@',
                regret_parameter='100_abs',
                model_group_ids=[1, 2],
                train_end_times=['2014-01-01', '2015-01-01'],
                plot_type='metric'
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert 'raw_value_next_time' in kwargs['frame']
        assert 'train_end_time' in kwargs['frame']
        assert kwargs['x_col'] == 'train_end_time'
        assert kwargs['y_col'] == 'raw_value_next_time'



import logging


from triage.component.catwalk.predictors import Predictor
from triage.component.catwalk.individual_importance import IndividualImportanceCalculator
from triage.component.catwalk.evaluation import ModelEvaluator


class ModelTester(object):
    def __init__(self, db_engine, model_storage_engine, matrix_storage_engine, replace, evaluator_config, individual_importance_config):
        self.matrix_storage_engine = matrix_storage_engine
        self.predictor = Predictor(
            db_engine=db_engine,
            model_storage_engine=model_storage_engine,
            replace=replace
        )

        self.individual_importance_calculator = IndividualImportanceCalculator(
            db_engine=db_engine,
            n_ranks=individual_importance_config.get('n_ranks', 5),
            methods=individual_importance_config.get('methods', ['uniform']),
            replace=replace
        )

        self.evaluator = ModelEvaluator(
            db_engine=db_engine,
            sort_seed=evaluator_config.get('sort_seed', None),
            testing_metric_groups=evaluator_config.get('testing_metric_groups', []),
            training_metric_groups=evaluator_config.get('training_metric_groups', [])
        )

    def generate_model_test_tasks(self, split, train_store, model_ids):
        test_tasks = []
        for test_matrix_def, test_uuid in zip(
            split['test_matrices'],
            split['test_uuids']
        ):
            test_store = self.matrix_storage_engine.get_store(test_uuid)

            if test_store.empty:
                logging.warning('''Test matrix for uuid %s
                was empty, no point in generating predictions. Not creating test task.
                ''', test_uuid)
                continue
            test_tasks.append({
                'test_store': test_store,
                'train_store': train_store,
                'model_ids': [model_id for model_id in model_ids if model_id]
            })
        return test_tasks

    def process_model_test_task(self, test_store, train_store, model_ids):
        as_of_times = test_store.metadata['as_of_times']
        logging.info(
            'Testing and scoring all model ids with test matrix %s. as_of_times min: %s max: %s num: %s',
            test_store.uuid,
            min(as_of_times),
            max(as_of_times),
            len(as_of_times)
        )
        for model_id in model_ids:
            logging.info('Testing model id %s', model_id)

            self.individual_importance_calculator\
                .calculate_and_save_all_methods_and_dates(
                    model_id,
                    test_store
                )

            # Generate predictions for the testing data then training data
            for store in (test_store, train_store):
                predictions_proba = self.predictor.predict(
                    model_id,
                    store,
                    misc_db_parameters=dict(),
                    train_matrix_columns=train_store.columns()
                )

                self.evaluator.evaluate(
                    predictions_proba=predictions_proba,
                    matrix_store=store,
                    model_id=model_id,
                )



import logging
from triage.component.architect.feature_group_creator import FeatureGroup


def leave_one_in(feature_groups):
    """For each group, return a copy of just that group

    Args:
        feature_groups (list) The feature groups to apply the strategy to

    Returns: A list of feature dicts
    """
    return feature_groups


def leave_one_out(feature_groups):
    """For each group, return a copy of all groups excluding that group

    Args:
        feature_groups (list) The feature groups to apply the strategy to

    Returns: A list of feature dicts
    """
    results = []
    for index_to_exclude in range(0, len(feature_groups)):
        group_copy = feature_groups.copy()
        del group_copy[index_to_exclude]
        feature_dict = FeatureGroup()
        for group in group_copy:
            feature_dict.update(group)
        results.append(feature_dict)
    return results


def all_features(feature_groups):
    """Return a combination of all feature groups

    Args:
        feature_groups (list) The feature groups to apply the strategy to

    Returns: A list of feature dicts
    """
    feature_dict = FeatureGroup()
    for group in feature_groups:
        feature_dict.update(group)
    return [feature_dict]


class FeatureGroupMixer(object):
    """Generates different combinations of feature groups
    based on a list of strategies"""
    strategy_lookup = {
        'leave-one-out': leave_one_out,
        'leave-one-in': leave_one_in,
        'all': all_features,
    }

    def __init__(self, strategies):
        for strategy in strategies:
            if strategy not in self.strategy_lookup:
                raise ValueError('Unknown strategy "{}"'.format(strategy))
        self.strategies = strategies

    def generate(self, feature_groups):
        """Apply all strategies to the list of feature groups

        Args:
            feature_groups (list) A list of feature dicts,
                each representing a group
        Returns: (list) of feature dicts
        """
        final_results = []
        for strategy in self.strategies:
            logging.info('Mixing feature groups %s using strategy %s', feature_groups, strategy)
            results = self.strategy_lookup[strategy](feature_groups)
            logging.info('Mixing found new feature groups %s', results)
            final_results += results

        return final_results



import logging
from abc import ABC, abstractmethod

from triage.component.architect.database_reflection import table_has_data
from triage.database_reflection import table_row_count


DEFAULT_ACTIVE_STATE = 'active'


class StateTableGeneratorBase(ABC):
    """Create a table containing the state of entities on different dates

    Referred to as a 'sparse states table', this table contains
        entities, dates, and boolean values for different states.

        These types of tables rarely exist in source data, but are useful
        internally in pipelines to express state logic in SQL

        For instance: '(in_jail OR booked) AND in_mental_health_facility'

        The output format is entity id/date/state1/state3/state3...

    Subclasses must implement the following methods:
        '_create_and_populate_sparse_table' to take dates
            and return a query to create the states table for those dates
        '_empty_table_message' to provide a helpful message to the user
            if no rows are found in the resultant table

    The main interface of StateTableGenerator objects is the
    `generate_sparse_table` method, which produces the latter
    'sparse'-style table.

    Args:
        db_engine (sqlalchemy.engine)
        experiment_hash (string) unique identifier for the experiment

    """
    def __init__(self, db_engine, experiment_hash):
        self.db_engine = db_engine
        self.experiment_hash = experiment_hash

    @abstractmethod
    def _create_and_populate_sparse_table(self, as_of_dates):
        pass

    @abstractmethod
    def _empty_table_message(self, as_of_dates):
        pass

    @property
    def sparse_table_name(self):
        return 'tmp_sparse_states_{}'.format(self.experiment_hash)

    def generate_sparse_table(self, as_of_dates):
        """Convert the object's input table
        into a sparse states table for the given as_of_dates

        Args:
            as_of_dates (list of datetime.dates) Dates to include in the sparse
                state table
        """
        logging.debug('Generating sparse table using as_of_dates: %s', as_of_dates)
        self._create_and_populate_sparse_table(as_of_dates)
        self.db_engine.execute(
            'create index on {} (entity_id, as_of_date)'
            .format(self.sparse_table_name)
        )
        logging.info('Indices created on entity_id and as_of_date for sparse state table')
        if not table_has_data(self.sparse_table_name, self.db_engine):
            raise ValueError(self._empty_table_message(as_of_dates))

        logging.info('Sparse states table generated at %s', self.sparse_table_name)
        logging.info('Generating stats on %s', self.sparse_table_name)
        logging.info('Row count of %s: %s',
                     self.sparse_table_name,
                     table_row_count(self.sparse_table_name, self.db_engine))

    def clean_up(self):
        self.db_engine.execute(
            'drop table if exists {}'.format(self.sparse_table_name)
        )


class StateTableGeneratorFromEntities(StateTableGeneratorBase):
    """Generates a 'sparse'-style states table from a table containing entity_ids

    This will include all entities found in the table for all given dates

    Args:
        entities_table (string, optional) name of SQL table containing
           entities
    """

    def __init__(self, entities_table, *args, **kwargs):

        super(StateTableGeneratorFromEntities, self).__init__(*args, **kwargs)
        self.entities_table = entities_table

    def _create_and_populate_sparse_table(self, as_of_dates):
        """Create a 'sparse'-style table from an entities table
            for a specific set of dates

        This will include all entities for all given dates

        Args:
        as_of_dates (list of datetime.date): Dates to calculate entity states as of
        """

        query = '''
            create table {sparse_state_table} as (
            select e.entity_id, a.as_of_date::timestamp, true {active_state}
                from {entities_table} e
                cross join (select unnest(ARRAY{as_of_dates}) as as_of_date) a
                group by e.entity_id, a.as_of_date
            )
        '''.format(
            sparse_state_table=self.sparse_table_name,
            entities_table=self.entities_table,
            as_of_dates=[date.isoformat() for date in as_of_dates],
            active_state=DEFAULT_ACTIVE_STATE
        )
        logging.debug('Assembled sparse state table query: %s', query)
        self.db_engine.execute(query)

    def _empty_table_message(self, as_of_dates):
        return "No entities in entities table '{input_table}'".format(
            input_table=self.entities_table,
        )


class StateTableGeneratorFromQuery(StateTableGeneratorBase):
    """Generates a 'sparse'-style states table from a given query

    Args:
    query (string) SQL query string to select entities for a given as_of_date
        The as_of_date should be parameterized with brackets: {as_of_date}
    """

    def __init__(self, query, *args, **kwargs):

        super(StateTableGeneratorFromQuery, self).__init__(*args, **kwargs)
        self.query = query

    def _create_and_populate_sparse_table(self, as_of_dates):
        """Create a 'sparse'-style states table by sequentially running a
            given date-parameterized query for all known dates.

        Args:
        as_of_dates (list of datetime.date): Dates to calculate entity states as of
        """

        self.db_engine.execute(
            f'''create table {self.sparse_table_name} (
                entity_id integer,
                as_of_date timestamp,
                {DEFAULT_ACTIVE_STATE} boolean
            )
            '''
        )
        logging.info('Created sparse state table, now inserting rows')

        for as_of_date in as_of_dates:
            formatted_date = f"{as_of_date.isoformat()}"
            dated_query = self.query.replace('{as_of_date}', formatted_date)
            full_query = f'''insert into {self.sparse_table_name}
                select q.entity_id, '{formatted_date}'::timestamp, true
                from ({dated_query}) q
                group by 1, 2, 3
            '''
            logging.info(
                f'Running state query for date: {as_of_date}, {full_query}',
            )
            self.db_engine.execute(full_query)

    def _empty_table_message(self, as_of_dates):
        return """Query does not return any rows for the given as_of_dates:
            {as_of_dates}
            '{query}'""".format(
            query=self.query,
            as_of_dates=', '.join(str(as_of_date) for as_of_date in (
                as_of_dates if len(as_of_dates) <= 5 else as_of_dates[:5] + ['…']
            )),
        )


class StateTableGeneratorFromDense(StateTableGeneratorBase):
    """Creates a 'sparse'-style states table from a 'dense'-style states table.

    A dense states table contains entities and *time ranges* for different states

    An example of this would be a building permits table, with
    time ranges that buildings are permitted.

    The expected format is entity id/state/start/end

    Args:
        dense_state_table (string, optional) name of SQL table containing
            entities and state time ranges
    """
    def __init__(self, dense_state_table, *args, **kwargs):
        super(StateTableGeneratorFromDense, self).__init__(*args, **kwargs)
        self.dense_state_table = dense_state_table

    def all_known_states(self):
        all_states = [
            row[0] for row in
            self.db_engine.execute('''
                select distinct(state) from {} order by state
            '''.format(self.dense_state_table))
        ]
        logging.info('Distinct states found: %s', all_states)
        return all_states

    def state_columns(self):
        state_columns = [
            "bool_or(state = '{desired_state}') as {desired_state}"
            .format(desired_state=state)
            for state in self.all_known_states()
        ]
        if not state_columns:
            raise ValueError("Unable to identify states from table",
                             self.dense_state_table)
        return state_columns

    def _create_and_populate_sparse_table(self, as_of_dates):
        """Creates a sparse states table from a dense states table for a given list of dates

        Args:
        as_of_dates (list of datetime.date): Dates to calculate entity states as of

        Returns: (string) A query to produce a sparse states table
        """
        query = '''
            create table {sparse_state_table} as (
            select d.entity_id, a.as_of_date::timestamp, {state_column_string}
                from {dense_state_table} d
                join (select unnest(ARRAY{as_of_dates}) as as_of_date) a
                on (
                    d.start_time <= a.as_of_date::timestamp and
                    d.end_time > a.as_of_date::timestamp
                )
                group by d.entity_id, a.as_of_date
            )
        '''.format(
            sparse_state_table=self.sparse_table_name,
            dense_state_table=self.dense_state_table,
            as_of_dates=[date.isoformat() for date in as_of_dates],
            state_column_string=', '.join(self.state_columns())
        )
        logging.debug('Assembled sparse state table query: %s', query)
        self.db_engine.execute(query)

    def _empty_table_message(self, as_of_dates):
        return \
            "No entities in dense state table '{input_table}' define time ranges " + \
            "that encompass any of experiment's \"as-of-dates\":\n\n" + \
            "\t{as_of_dates}\n\n" + \
            "Please check temporal config and dense state table".format(
                input_table=self.dense_state_table,
                as_of_dates=', '.join(str(as_of_date) for as_of_date in (
                    as_of_dates if len(as_of_dates) <= 5 else as_of_dates[:5] + ['…']
                )),
            )


class StateTableGeneratorNoOp(object):
    def generate_sparse_table(self, as_of_dates):
        logging.warning('No cohort configuration is available, so no cohort will be created')
        return

    def clean_up(self):
        logging.warning('No cohort table exists, so nothing to tear down')
        return

    @property
    def sparse_table_name(self):
        return None



import testing.postgresql
from sqlalchemy import create_engine

from triage.component.results_schema import Base

from .factories import (
    ModelGroupFactory,
    ModelFactory,
    EvaluationFactory,
    PredictionFactory,
    IndividualImportanceFactory,
    init_engine,
    session,
    MatrixFactory
)


def test_evaluation_factories():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        Base.metadata.create_all(engine)
        init_engine(engine)

        model_group = ModelGroupFactory()
        model = ModelFactory(model_group_rel=model_group)
        for metric, value in [
            ('precision@', 0.4),
            ('recall@', 0.3),
        ]:
            EvaluationFactory(
                model_rel=model,
                metric=metric,
                parameter='100_abs',
                value=value
            )
        session.commit()
        results = engine.execute('''
            select
                model_group_id,
                m.model_id,
                e.metric,
                e.value
            from
                test_results.evaluations e
                join model_metadata.models m using (model_id)
        ''')
        for model_group_id, model_id, metric, value in results:
            # if the evaluations are created with the model group and model,
            # as opposed to an autoprovisioned one,
            # the ids in a fresh DB should be 1
            assert model_group_id == 1
            assert model_id == 1


def test_prediction_factories():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        Base.metadata.create_all(engine)
        init_engine(engine)

        # create some basic predictions, but with the same model group and
        # model to test the factory relationships
        model_group = ModelGroupFactory()
        model = ModelFactory(model_group_rel=model_group)
        for entity_id, as_of_date in [
            (1, '2016-01-01'),
            (1, '2016-04-01'),
            (2, '2016-01-01'),
            (2, '2016-04-01'),
            (3, '2016-01-01'),
            (3, '2016-04-01'),
        ]:
            PredictionFactory(
                model_rel=model,
                entity_id=entity_id,
                as_of_date=as_of_date,
            )
            IndividualImportanceFactory(
                model_rel=model,
                entity_id=entity_id,
                as_of_date=as_of_date,
            )
        session.commit()
        results = engine.execute('''
            select m.*, p.*
            from
                test_results.predictions p
                join model_metadata.models m using (model_id)
                join test_results.individual_importances i using (model_id, entity_id, as_of_date)
        ''')
        assert len([row for row in results]) == 6
        # if the predictions are created with the model,
        # the join should work and we should have the original six results



from unittest.mock import patch

import pandas
from matplotlib import lines as mlines

from triage.component.audition.plotting import (
    generate_plot_lines,
    category_colordict,
    category_styledict,
    plot_cats,
)


def test_generate_plot_lines():
    colordict = {
        'cat1': '#001122',
        'cat2': '#112233',
        'cat3': '#223344',
    }
    styledict = {
        'cat1': '-',
        'cat2': '--',
        'cat3': '-',
    }
    plot_lines = generate_plot_lines(colordict,
                                     lambda x: 'Cat {}'.format(x),
                                     styledict)
    assert len(plot_lines) == 3
    for line in plot_lines:
        assert type(line) == mlines.Line2D
        assert 'Cat ' in line._label
        assert '-' in line._linestyle
        if line._label == 'Cat 2':
            assert line._linestyle == '--'


def test_category_colordict():
    cmap_name = 'Vega10'
    categories = ['Cat1', 'Cat2', 'Cat3', 'Cat4']
    colordict = category_colordict(cmap_name, categories)
    assert len(colordict.keys()) == 4


def test_category_colordict_with_highlight():
    cmap_name = 'Vega10'
    colordict_with_highlight = category_colordict(
        cmap_name,
        ['Cat1', 'Cat2', 'Cat3', 'Cat4'],
        'Cat2'
    )
    colordict_without_highlight = category_colordict(
        cmap_name,
        ['Cat1', 'Cat3', 'Cat4']
    )
    for cat in ['Cat1', 'Cat3', 'Cat4']:
        assert colordict_with_highlight[cat] == colordict_without_highlight[cat]
    assert colordict_with_highlight['Cat2'] == '#000000'


def test_category_styledict():
    colordict = {
        'cat1': '#001122',
        'cat2': '#112233',
        'cat3': '#223344',
    }
    assert category_styledict(colordict, 'cat3') == {
        'cat1': '-',
        'cat2': '-',
        'cat3': '--',
    }


def test_plot_cats():
    test_df = pandas.DataFrame.from_dict({
        'cats': ['tuxedo', 'maine coon', 'lion!'],
        'groups': ['i', 'dont', 'know'],
        'col1': [1, 2, 3],
        'col2': [4, 5, 6],
        'col3': [7, 8, 9],
    })
    # hard to make many assertions, but we can make sure it gets to the end
    # and shows the contents
    with patch('triage.component.audition.plotting.plt.show') as show_patch:
        plot_cats(test_df, 'col1', 'col2', cat_col='cats', grp_col='groups')
        assert show_patch.called



import os.path

from sqlalchemy import (
    Column,
    BigInteger,
    Boolean,
    Integer,
    Interval,
    String,
    Numeric,
    DateTime,
    JSON,
    Float,
    Text,
    ForeignKey,
    MetaData,
    DDL,
    event
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.types import ARRAY
from sqlalchemy.sql import func

# One declarative_base object for each schema created
Base = declarative_base()

schemas = ("CREATE SCHEMA IF NOT EXISTS model_metadata;"
           " CREATE SCHEMA IF NOT EXISTS test_results;"
           " CREATE SCHEMA IF NOT EXISTS train_results;")

event.listen(
    Base.metadata,
    'before_create',
    DDL(schemas)
)

group_proc_filename = os.path.join(
    os.path.dirname(__file__),
    'model_group_stored_procedure.sql'
)
with open(group_proc_filename) as fd:
    stmt = fd.read()

event.listen(
    Base.metadata,
    'before_create',
    DDL(stmt)
)


class Experiment(Base):

    __tablename__ = 'experiments'
    __table_args__ = {"schema": "model_metadata"}

    experiment_hash = Column(String, primary_key=True)
    config = Column(JSONB)


class ModelGroup(Base):

    __tablename__ = 'model_groups'
    __table_args__ = {"schema": "model_metadata"}

    model_group_id = Column(Integer, primary_key=True)
    model_type = Column(Text)
    hyperparameters = Column(JSONB)
    feature_list = Column(ARRAY(Text))
    model_config = Column(JSONB)


class ListPrediction(Base):

    __tablename__ = 'list_predictions'
    __table_args__ = {"schema": "model_metadata"}

    model_id = Column(Integer, ForeignKey('model_metadata.models.model_id'), primary_key=True)
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    score = Column(Numeric)
    rank_abs = Column(Integer)
    rank_pct = Column(Float)
    matrix_uuid = Column(Text)
    test_label_timespan = Column(Interval)

    model_rel = relationship('Model')


class Matrix(Base):

    __tablename__ = 'matrices'
    __table_args__ = {"schema": "model_metadata"}

    matrix_id = Column(String)
    matrix_uuid = Column(String, unique=True, index=True, primary_key=True)
    matrix_type = Column(String) # 'train' or 'test'
    labeling_window = Column(Interval)
    num_observations = Column(Integer)
    creation_time = Column(DateTime(timezone=True), server_default=func.now())
    lookback_duration = Column(Interval)
    feature_start_time = Column(DateTime)
    matrix_metadata = Column(JSONB)



class Model(Base):

    __tablename__ = 'models'
    __table_args__ = {"schema": "model_metadata"}

    model_id = Column(Integer, primary_key=True)
    model_group_id = Column(Integer, ForeignKey('model_metadata.model_groups.model_group_id'))
    model_hash = Column(String, unique=True, index=True)
    run_time = Column(DateTime)
    batch_run_time = Column(DateTime)
    model_type = Column(String)
    hyperparameters = Column(JSONB)
    model_comment = Column(Text)
    batch_comment = Column(Text)
    config = Column(JSON)
    experiment_hash = Column(String, ForeignKey('model_metadata.experiments.experiment_hash'))
    train_end_time = Column(DateTime)
    test = Column(Boolean)
    train_matrix_uuid = Column(Text, ForeignKey('model_metadata.matrices.matrix_uuid'))
    training_label_timespan = Column(Interval)
    model_size = Column(Float)

    model_group_rel = relationship('ModelGroup')
    experiment_rel = relationship('Experiment')
    matrix_rel = relationship('Matrix')

    def delete(self, session):
        # basically implement a cascade, in case cascade is not implemented
        (session.query(FeatureImportance)
            .filter_by(model_id=self.model_id)
            .delete())
        (session.query(Evaluation)
            .filter_by(model_id=self.model_id)
            .delete())
        (session.query(Prediction)
            .filter_by(model_id=self.model_id)
            .delete())
        session.delete(self)


class FeatureImportance(Base):

    __tablename__ = 'feature_importances'
    __table_args__ = {"schema": "train_results"}

    model_id = Column(Integer, ForeignKey('model_metadata.models.model_id'), primary_key=True)
    model = relationship(Model)
    feature = Column(String, primary_key=True)
    feature_importance = Column(Numeric)
    rank_abs = Column(Integer)
    rank_pct = Column(Float)

    model_rel = relationship('Model')


class TestPrediction(Base):

    __tablename__ = 'predictions'
    __table_args__ = {"schema": "test_results"}

    model_id = Column(Integer, ForeignKey('model_metadata.models.model_id'), primary_key=True)
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    score = Column(Numeric)
    label_value = Column(Integer)
    rank_abs = Column(Integer)
    rank_pct = Column(Float)
    matrix_uuid = Column(Text, ForeignKey('model_metadata.matrices.matrix_uuid'))
    test_label_timespan = Column(Interval)

    model_rel = relationship('Model')

class TrainPrediction(Base):

    __tablename__ = 'predictions'
    __table_args__ = {"schema": "train_results"}

    model_id = Column(Integer, ForeignKey('model_metadata.models.model_id'), primary_key=True)
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    score = Column(Numeric)
    label_value = Column(Integer)
    rank_abs = Column(Integer)
    rank_pct = Column(Float)
    matrix_uuid = Column(Text, ForeignKey('model_metadata.matrices.matrix_uuid'))
    test_label_timespan = Column(Interval)

    model_rel = relationship('Model')


class IndividualImportance(Base):

    __tablename__ = 'individual_importances'
    __table_args__ = {"schema": "test_results"}

    model_id = Column(Integer, ForeignKey('model_metadata.models.model_id'), primary_key=True)
    entity_id = Column(BigInteger, primary_key=True)
    as_of_date = Column(DateTime, primary_key=True)
    feature = Column(String, primary_key=True)
    method = Column(String, primary_key=True)
    feature_value = Column(Float)
    importance_score = Column(Float)

    model_rel = relationship('Model')


class TestEvaluation(Base):

    __tablename__ = 'evaluations'
    __table_args__ = {"schema": "test_results"}

    model_id = Column(Integer, ForeignKey('model_metadata.models.model_id'), primary_key=True)
    evaluation_start_time = Column(DateTime, primary_key=True)
    evaluation_end_time = Column(DateTime, primary_key=True)
    as_of_date_frequency = Column(Interval, primary_key=True)
    metric = Column(String, primary_key=True)
    parameter = Column(String, primary_key=True)
    value = Column(Numeric)
    num_labeled_examples = Column(Integer)
    num_labeled_above_threshold = Column(Integer)
    num_positive_labels = Column(Integer)
    sort_seed = Column(Integer)

    model_rel = relationship('Model')


class TrainEvaluation(Base):

    __tablename__ = 'evaluations'
    __table_args__ = {"schema": "train_results"}

    model_id = Column(Integer, ForeignKey('model_metadata.models.model_id'), primary_key=True)
    evaluation_start_time = Column(DateTime, primary_key=True)
    evaluation_end_time = Column(DateTime, primary_key=True)
    as_of_date_frequency = Column(Interval, primary_key=True)
    metric = Column(String, primary_key=True)
    parameter = Column(String, primary_key=True)
    value = Column(Numeric)
    num_labeled_examples = Column(Integer)
    num_labeled_above_threshold = Column(Integer)
    num_positive_labels = Column(Integer)
    sort_seed = Column(Integer)

    model_rel = relationship('Model')



import logging
import traceback
from functools import partial
from multiprocessing import Pool

from triage.component.catwalk.utils import Batch

from triage.experiments import ExperimentBase


class MultiCoreExperiment(ExperimentBase):
    def __init__(self, n_processes=1, n_db_processes=1, *args, **kwargs):
        super(MultiCoreExperiment, self).__init__(*args, **kwargs)
        if n_processes < 1:
            raise ValueError('n_processes must be 1 or greater')
        if n_db_processes < 1:
            raise ValueError('n_db_processes must be 1 or greater')
        if n_db_processes == 1 and n_processes == 1:
            logging.warning('Both n_processes and n_db_processes were set to 1. '
                            'If you only wish to use one process to run the experiment, '
                            'consider using the SingleThreadedExperiment class instead')
        self.n_processes = n_processes
        self.n_db_processes = n_db_processes

    def generated_chunked_parallelized_results(
        self,
        partially_bound_function,
        tasks,
        n_processes,
        chunksize=1,
    ):
        with Pool(n_processes, maxtasksperchild=1) as pool:
            for result in pool.map(
                partially_bound_function,
                [list(task_batch) for task_batch in Batch(tasks, chunksize)]
            ):
                yield result

    def process_train_tasks(self, train_tasks):
        partial_train_models = partial(
            run_task_with_splatted_arguments,
            self.trainer.process_train_task
        )
        logging.info(
            'Starting parallel training: %s tasks, %s processes',
            len(train_tasks),
            self.n_processes
        )
        model_ids = []
        for model_id in parallelize(
            partial_train_models,
            train_tasks,
            self.n_processes
        ):
            model_ids.append(model_id)
        return model_ids

    def process_model_test_tasks(self, test_tasks):
        partial_test = partial(
            run_task_with_splatted_arguments,
            self.tester.process_model_test_task
        )

        logging.info(
            'Starting parallel testing with %s processes',
            self.n_db_processes
        )
        parallelize(
            partial_test,
            test_tasks,
            self.n_db_processes
        )
        logging.info('Cleaned up concurrent pool')

    def process_query_tasks(self, query_tasks):
        logging.info(
            'Processing query tasks with %s processes',
            self.n_db_processes
        )
        for table_name, tasks in query_tasks.items():
            logging.info('Processing features for %s', table_name)
            self.feature_generator.run_commands(tasks.get('prepare', []))
            partial_insert = partial(insert_into_table, feature_generator=self.feature_generator)

            insert_batches = [list(task_batch) for task_batch in Batch(tasks.get('inserts', []), 25)]
            parallelize(
                partial_insert,
                insert_batches,
                n_processes=self.n_db_processes,
            )
            self.feature_generator.run_commands(tasks.get('finalize', []))
            logging.info('%s completed', table_name)

    def process_matrix_build_tasks(self, matrix_build_tasks):
        partial_build_matrix = partial(
            run_task_with_splatted_arguments,
            self.matrix_builder.build_matrix
        )
        logging.info(
            'Starting parallel matrix building: %s matrices, %s processes',
            len(self.matrix_build_tasks.keys()),
            self.n_processes
        )
        parallelize(
            partial_build_matrix,
            self.matrix_build_tasks.values(),
            self.n_processes
        )


def insert_into_table(insert_statements, feature_generator):
    try:
        logging.info('Beginning insert batch')
        feature_generator.run_commands(insert_statements)
        return True
    except Exception:
        logging.error('Child error: %s', traceback.format_exc())
        return False


def parallelize(partially_bound_function, tasks, n_processes):
    num_successes = 0
    num_failures = 0
    results = []
    with Pool(n_processes, maxtasksperchild=1) as pool:
        for result in pool.map(partially_bound_function, tasks):
            if result:
                num_successes += 1
            else:
                num_failures += 1
            results.append(result)

        logging.info(
            'Done. successes: %s, failures: %s',
            num_successes,
            num_failures
        )
        return results


def run_task_with_splatted_arguments(task_runner, task):
    try:
        return task_runner(**task)
    except Exception:
        logging.error('Child error: %s', traceback.format_exc())
        return None



"""Main application"""
from .model_trainers import ModelTrainer
from .predictors import Predictor
from .evaluation import ModelEvaluator

__all__ = (
    'ModelTrainer',
    'Predictor',
    'ModelEvaluator',
)



# coding: utf-8

from io import BytesIO
import os
from os.path import dirname
import pathlib
import logging
from sklearn.externals import joblib
from urllib.parse import urlparse
from triage.component.results_schema import TestEvaluation, TrainEvaluation, \
    TestPrediction, TrainPrediction

import pandas as pd
import s3fs
import yaml


class Store(object):
    """Base class for classes which know how to access a file in a preset medium.

    Used to hold references to persisted objects with knowledge about how they can be accessed.
    without loading them into memory. In this way, they can be easily and quickly serialized
    across processes but centralize the reading/writing code.

    Each subclass be scoped to a specific storage medium (e.g. Filesystem, S3)
        and implement the access methods for that medium.

    Implements write/load methods for interacting directly using bytestreams,
        plus an open method that works as an open filehandle.
    """
    def __init__(self, *pathparts):
        self.pathparts = pathparts

    @classmethod
    def factory(self, *pathparts):
        path_parsed = urlparse(pathparts[0])
        scheme = path_parsed.scheme

        if scheme in ('', 'file'):
            return FSStore(*pathparts)
        elif scheme == 's3':
            return S3Store(*pathparts)
        else:
            raise ValueError('Unable to infer correct Store from project path')

    def __str__(self):
        return f"{self.__class__.__name__}(path={self.path})"

    def __repr__(self):
        return str(self)

    def exists(self):
        raise NotImplementedError

    def load(self):
        with self.open('rb') as fd:
            return fd.read()

    def write(self, bytestream):
        with self.open('wb') as fd:
            fd.write(bytestream)

    def open(self, *args, **kwargs):
        raise NotImplementedError


class S3Store(Store):
    """Store an object in S3.
    
    Example:
    ```
    store = S3Store('s3://my-bucket', 'models', 'model.pkl')
    return store.load()
    ```

    Args:
        *pathparts: A variable length list of components of the path, to be processed in order.
            All components will be joined using PurePosixPath to create the final path.
    """
    def __init__(self, *pathparts):
        self.path = str(pathlib.PurePosixPath(pathparts[0].replace('s3://', ''), *pathparts[1:]))

    def exists(self):
        s3 = s3fs.S3FileSystem()
        return s3.exists(self.path)

    def delete(self):
        s3 = s3fs.S3FileSystem()
        s3.rm(self.path)

    def open(self, *args, **kwargs):
        s3 = s3fs.S3FileSystem()
        return s3.open(self.path, *args, **kwargs)


class FSStore(Store):
    """Store an object on the local filesystem.
    
    Example:
    ```
    store = FSStore('/mnt', 'models', 'model.pkl')
    return store.load()
    ```

    Args:
        *pathparts: A variable length list of components of the path, to be processed in order. 
            All components will be joined using pathlib.Path to create the final path
                using the correct separator for the operating system. However, if you pass
                components that already contain a separator, those separators won't be modified
    """
    def __init__(self, *pathparts):
        self.path = pathlib.Path(*pathparts)
        os.makedirs(dirname(self.path), exist_ok=True)

    def exists(self):
        return os.path.isfile(self.path)

    def delete(self):
        os.remove(self.path)

    def open(self, *args, **kwargs):
        return open(self.path, *args, **kwargs)


class ProjectStorage(object):
    """Store and access files associated with a project.
    
    Args:
        project_path (string): The base path for all files in the project.
            The scheme prefix of the path will determine the storage medium.
    """
    def __init__(self, project_path):
        self.project_path = project_path
        self.storage_class = Store.factory(self.project_path).__class__

    def get_store(self, directories, leaf_filename):
        """Return a storage object for one filename

        Args:
        directories (list): A list of subdirectories
        leaf_filename (string): The filename without any directory information

        Returns:
            triage.component.catwalk.storage.Store object
        """
        return self.storage_class(self.project_path, *directories, leaf_filename)
   
    def matrix_storage_engine(self, matrix_storage_class=None, matrix_directory=None):
        """Return a matrix storage engine bound to this project's storage

        Args:
            matrix_storage_class (class) A subclass of MatrixStore
            matrix_directory (string, optional) A directory to store matrices.
                If not passed will allow the MatrixStorageEngine to decide
        Returns: triage.component.catwalk.storage.MatrixStorageEngine
        """
        return MatrixStorageEngine(self, matrix_storage_class, matrix_directory)

    def model_storage_engine(self, model_directory=None):
        """Return a model storage engine bound to this project's storage

        Args:
            model_directory (string, optional) A directory to store models
                If not passed will allow the ModelStorageEngine to decide
        Returns: triage.component.catwalk.storage.ModelStorageEngine
        """
        return ModelStorageEngine(self, model_directory)


class ModelStorageEngine(object):
    """Store arbitrary models in a given project storage using joblib

    Args:
        project_storage (triage.component.catwalk.storage.ProjectStorage) A project file storage engine
        model_directory (string, optional) A directory name for models. Defaults to 'trained_models'
    
    """
    def __init__(self, project_storage, model_directory=None):
        self.project_storage = project_storage
        self.directories = [model_directory or 'trained_models']

    def write(self, obj, model_hash):
        """Persist a model object using joblib. Also performs compression

        Args:
            obj (object) A picklable model object
            model_hash (string) An identifier, unique within this project, for the model
        """
        with self._get_store(model_hash).open('wb') as fd:
            joblib.dump(obj, fd, compress=True)

    def load(self, model_hash):
        """Load a model object using joblib

        Args:
            model_hash (string) An identifier, unique within this project, for the model

        Returns: (object) A model object
        """
        with self._get_store(model_hash).open('rb') as fd:
            return joblib.load(fd)

    def exists(self, model_hash):
        """Check whether the model is persisted

        Args:
            model_hash (string) An identifier, unique within this project, for the model

        Returns: (bool) Whether or not a model by that identifier exists in project storage
        """
        return self._get_store(model_hash).exists()

    def delete(self, model_hash):
        """Delete the model identified by this hash from project storage

        Args:
            model_hash (string) An identifier, unique within this project, for the model
        """
        return self._get_store(model_hash).delete()

    def _get_store(self, model_hash):
        return self.project_storage.get_store(self.directories, model_hash)


class MatrixStorageEngine(object):
    """Store matrices in a given project storage

    Args:
        project_storage (triage.component.catwalk.storage.ProjectStorage) A project file storage engine
        matrix_storage_class (class) A subclass of MatrixStore
        matrix_directory (string, optional) A directory to store matrices. Defaults to 'matrices'
    """
    def __init__(self, project_storage, matrix_storage_class=None, matrix_directory=None):
        self.project_storage = project_storage
        self.matrix_storage_class = matrix_storage_class or CSVMatrixStore
        self.directories = [matrix_directory or 'matrices']

    def get_store(self, matrix_uuid):
        """Return a storage object for a given matrix uuid.

        Args:
            matrix_uuid (string) A unique identifier within the project for a matrix.

        Returns: (MatrixStore) a reference to the matrix and its companion metadata
        """
        return self.matrix_storage_class(self.project_storage, self.directories, matrix_uuid)


class MatrixStore(object):
    """Base class for classes that allow access of a matrix and its metadata.

    Subclasses should be scoped to a storage format (e.g. CSV, HDF)
        and implement the _load, save, and head_of_matrix methods for that storage format

    Args:
        project_storage (triage.component.catwalk.storage.ProjectStorage) A project file storage engine
        directories (list): A list of subdirectories
        matrix_uuid (string): A unique identifier within the project for a matrix.
        matrix (pandas.DataFrame, optional): The raw matrix. Defaults to None, which means it will be loaded from storage on demand
        metadata (dict, optional). The matrix' metadata. Defaults to None, which means it will be loaded from storage on demand.
    """
    _labels = None

    def __init__(self, project_storage, directories, matrix_uuid, matrix=None, metadata=None):
        self.matrix_uuid = matrix_uuid
        self.matrix_base_store = project_storage.get_store(directories, f'{matrix_uuid}.{self.suffix}')
        self.metadata_base_store = project_storage.get_store(directories, f'{matrix_uuid}.yaml')

        self.matrix = matrix
        self.metadata = metadata

    @property
    def matrix(self):
        """The raw matrix. Will load from storage into memory if not already loaded"""
        if self.__matrix is None:
            self.__matrix = self._load()
        return self.__matrix

    @matrix.setter
    def matrix(self, matrix):
        self.__matrix = matrix

    @property
    def metadata(self):
        """The raw metadata. Will load from storage into memory if not already loaded"""
        if self.__metadata is None:
            self.__metadata = self.load_metadata()
        return self.__metadata

    @metadata.setter
    def metadata(self, metadata):
        self.__metadata = metadata

    @property
    def head_of_matrix(self):
        """The first line of the matrix"""
        return self.matrix.head(1)

    @property
    def exists(self):
        """Whether or not the matrix and metadata exist in storage"""
        return self.matrix_base_store.exists() and self.metadata_base_store.exists()

    @property
    def empty(self):
        """Whether or not the matrix has at least one row"""
        if not self.matrix_base_store.exists():
            return True
        else:
            head_of_matrix = self.head_of_matrix
            return head_of_matrix.empty

    def columns(self, include_label=False):
        """The matrix's column list"""
        head_of_matrix = self.head_of_matrix
        columns = head_of_matrix.columns.tolist()
        if include_label:
            return columns
        else:
            return [
                col for col in columns
                if col != self.metadata['label_name']
            ]

    def labels(self):
        """The matrix's label column."""
        if self._labels is not None:
            logging.debug('using stored labels')
            return self._labels
        else:
            logging.debug('popping labels from matrix')
            self._labels = self.matrix.pop(self.metadata['label_name'])
            return self._labels

    @property
    def uuid(self):
        """The matrix's unique id within the project"""
        return self.matrix_uuid

    @property
    def as_of_dates(self):
        """The list of as-of-dates in the matrix"""
        if 'as_of_date' in self.matrix.index.names:
            return sorted(list(set([as_of_date for entity_id, as_of_date in self.matrix.index])))
        else:
            return [self.metadata['end_time']]

    @property
    def num_entities(self):
        """The number of entities in the matrix"""
        if self.matrix.index.names == ['entity_id']:
            return len(self.matrix.index.values)
        elif 'entity_id' in self.matrix.index.names:
            return len(self.matrix.index.levels[self.matrix.index.names.index('entity_id')])

    @property
    def matrix_type(self):
        """The MatrixType (train or test). Returns an object with:
            a string name,
            evaluation ORM class
            prediction ORM class
            a boolean `is_test`
        """
        if self.metadata['matrix_type'] == 'train':
            return TrainMatrixType
        elif self.metadata['matrix_type'] == 'test':
            return TestMatrixType
        else:
            raise Exception('''matrix metadata for matrix {} must contain 'matrix_type'
             = "train" or "test" '''.format(self.uuid))

    def matrix_with_sorted_columns(self, columns):
        """Return the matrix with columns sorted in the given column order

        Args:
            columns (list) The order of column names to return.
                Will error if this list does not contain the same elements as the matrix's columns
        """
        columnset = set(self.columns())
        desired_columnset = set(columns)
        if columnset == desired_columnset:
            if self.columns() != columns:
                logging.warning('Column orders not the same, re-ordering')
            return self.matrix[columns]
        else:
            if columnset.issuperset(desired_columnset):
                raise ValueError('''
                    Columnset is superset of desired columnset. Extra items: %s
                ''', columnset - desired_columnset)
            elif columnset.issubset(desired_columnset):
                raise ValueError('''
                    Columnset is subset of desired columnset. Extra items: %s
                ''', desired_columnset - columnset)
            else:
                raise ValueError('''
                    Columnset and desired columnset mismatch. Unique items: %s
                ''', columnset ^ desired_columnset)

    def load_metadata(self):
        """Load metadata from storage"""
        with self.metadata_base_store.open('rb') as fd:
            return yaml.load(fd)

    def save(self):
        raise NotImplementedError

    def __getstate__(self):
        """Remove object of a large size upon serialization. 

        This helps in a multiprocessing context.
        """
        self.matrix = None
        self._labels = None
        self.metadata = None
        return self.__dict__.copy()


class HDFMatrixStore(MatrixStore):
    """Store and access matrices using HDF"""
    suffix = 'h5'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.matrix_base_store, S3Store):
            raise ValueError('HDFMatrixStore cannot be used with S3')

    @property
    def head_of_matrix(self):
        try:
            head_of_matrix = pd.read_hdf(self.matrix_base_store.path, start=0, stop=1)
            # Is the index already in place?
            if head_of_matrix.index.names != self.metadata['indices']:
                head_of_matrix.set_index(self.metadata['indices'], inplace=True)
        except pd.errors.EmptyDataError:
            head_of_matrix = None

        return head_of_matrix

    def _load(self):
        matrix = pd.read_hdf(self.matrix_base_store.path)

        # Is the index already in place?
        if matrix.index.names != self.metadata['indices']:
            matrix.set_index(self.metadata['indices'], inplace=True)

        return matrix

    def save(self):
        hdf = pd.HDFStore(self.matrix_base_store.path,
                          mode='w',
                          complevel=4,
                          complib="zlib",
                          format='table')
        hdf.put(self.matrix_uuid, self.matrix.apply(pd.to_numeric), data_columns=True)
        hdf.close()

        with self.metadata_base_store.open('wb') as fd:
            yaml.dump(self.metadata, fd, encoding='utf-8')


class CSVMatrixStore(MatrixStore):
    """Store and access matrices using CSV"""
    suffix = 'csv'

    @property
    def head_of_matrix(self):
        try:
            with self.matrix_base_store.open('rb') as fd:
                head_of_matrix = pd.read_csv(fd, nrows=1)
                head_of_matrix.set_index(self.metadata['indices'], inplace=True)
        except FileNotFoundError as fnfe:
            logging.exception(f"Matrix isn't there: {fnfe}")
            logging.exception("Returning Empty data frame")
            head_of_matrix = pd.DataFrame()

        return head_of_matrix

    def _load(self):
        parse_dates_argument = ['as_of_date'] if 'as_of_date' in self.metadata['indices'] else False
        with self.matrix_base_store.open('rb') as fd:
            matrix = pd.read_csv(fd, parse_dates=parse_dates_argument)

        matrix.set_index(self.metadata['indices'], inplace=True)

        return matrix

    def save(self):
        self.matrix_base_store.write(self.matrix.to_csv(None).encode('utf-8'))
        with self.metadata_base_store.open('wb') as fd:
            yaml.dump(self.metadata, fd, encoding='utf-8')


class TestMatrixType(object):
    string_name = 'test'
    evaluation_obj = TestEvaluation
    prediction_obj = TestPrediction
    is_test = True


class TrainMatrixType(object):
    string_name = 'train'
    evaluation_obj = TrainEvaluation
    prediction_obj = TrainPrediction
    is_test = False



import itertools

from triage.component.architect.feature_group_mixer import FeatureGroupMixer
from triage.component.architect.feature_group_creator import FeatureGroup

import pytest


@pytest.fixture
def english_numbers():
    return FeatureGroup(
        name='english_numbers',
        features_by_table={'one': ['two', 'three'], 'four': ['five', 'six']}
    )


@pytest.fixture
def letters():
    return FeatureGroup(
        name='letters',
        features_by_table={'a': ['b', 'c'], 'd': ['e', 'f']}
    )


@pytest.fixture
def german_numbers():
    return FeatureGroup(
        name='german_numbers',
        features_by_table={'eins': ['zwei', 'drei'], 'vier': ['funf', 'sechs']}
    )


def test_feature_group_mixer_leave_one_out(english_numbers, letters, german_numbers):
    feature_groups = [english_numbers, letters, german_numbers]

    result = FeatureGroupMixer(['leave-one-out']).generate(feature_groups)
    expected = [
        dict(itertools.chain(letters.items(), german_numbers.items())),
        dict(itertools.chain(english_numbers.items(), german_numbers.items())),
        dict(itertools.chain(english_numbers.items(), letters.items())),
    ]
    assert result == expected
    assert [g.names for g in result] == [
        ['letters', 'german_numbers'],
        ['english_numbers', 'german_numbers'],
        ['english_numbers', 'letters'],
    ]


def test_feature_group_mixer_leave_one_in(english_numbers, letters, german_numbers):
    feature_groups = [english_numbers, letters, german_numbers]

    result = FeatureGroupMixer(['leave-one-in']).generate(feature_groups)
    expected = [
        english_numbers,
        letters,
        german_numbers
    ]
    assert result == expected
    assert [g.names for g in result] == [
        ['english_numbers'],
        ['letters'],
        ['german_numbers'],
    ]


def test_feature_group_mixer_all(english_numbers, letters, german_numbers):
    feature_groups = [english_numbers, letters, german_numbers]

    result = FeatureGroupMixer(['all']).generate(feature_groups)
    expected = [
        dict(itertools.chain(english_numbers.items(), letters.items(), german_numbers.items())),
    ]
    assert result == expected
    assert result[0].names == ['english_numbers', 'letters', 'german_numbers']



from sqlalchemy import create_engine
import testing.postgresql

from triage.component.catwalk.db import ensure_db

from tests.utils import sample_config, populate_source_data
from triage.experiments.validate import ExperimentValidator


def test_experiment_validator():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        populate_source_data(db_engine)
        ExperimentValidator(db_engine).run(sample_config())



import copy
import itertools
import logging

from triage.component import metta

from . import utils, state_table_generators


class Planner(object):

    def __init__(
        self,
        feature_start_time,
        label_names,
        label_types,
        states,
        user_metadata,
        cohort_name='default',
    ):
        self.feature_start_time = feature_start_time  # earliest time included in features
        self.label_names = label_names
        self.label_types = label_types
        self.cohort_name = cohort_name
        self.states = states or [state_table_generators.DEFAULT_ACTIVE_STATE]
        self.user_metadata = user_metadata

    def _generate_build_task(
        self,
        matrix_metadata,
        matrix_uuid,
        train_matrix,
        feature_dictionary
    ):
        return {
            'as_of_times': train_matrix['as_of_times'],
            'label_name': matrix_metadata['label_name'],
            'label_type': matrix_metadata['label_type'],
            'feature_dictionary': feature_dictionary,
            'matrix_uuid': matrix_uuid,
            'matrix_metadata': matrix_metadata,
            'matrix_type': matrix_metadata['matrix_type']
        }

    def _make_metadata(self, matrix_definition, feature_dictionary, label_name,
                       label_type, state, matrix_type):
        """ Generate dictionary of matrix metadata.

        :param matrix_definition: temporal definition of matrix
        :param feature dictionary: feature tables and the columns within them to use as features
        :param label_name: name of label column
        :param label_type: type of label
        :param state: the entity state to be included in the matrix
        :param matrix_type: type (train/test) of matrix
        :type matrix_definition: dict
        :type feature dictionary: dict
        :type label_name: str
        :type label_type: str
        :type state: str
        :type matrix_type: str

        :return: metadata needed for matrix identification and modeling
        :rtype: dict
        """

        # make a human-readable label for this matrix
        matrix_id = '_'.join([
            label_name,
            label_type,
            str(matrix_definition['first_as_of_time']),
            str(matrix_definition['matrix_info_end_time'])
        ])
        matrix_metadata = {

            # temporal information
            'feature_start_time': self.feature_start_time,
            'end_time': matrix_definition['matrix_info_end_time'],
            'as_of_date_frequency': matrix_definition.get(
                'training_as_of_date_frequency',
                matrix_definition.get('test_as_of_date_frequency')
            ),

            # columns
            'indices': ['entity_id', 'as_of_date'],
            'feature_names': utils.feature_list(feature_dictionary),
            'feature_groups': feature_dictionary.names,
            'label_name': label_name,

            # other information
            'label_type': label_type,
            'label_timespan': matrix_definition.get(
                'test_label_timespan',
                matrix_definition.get('training_label_timespan', '0 days')
            ),
            'cohort_name': self.cohort_name,
            'state': state,
            'matrix_id': matrix_id,
            'matrix_type': matrix_type

        }
        matrix_metadata.update(matrix_definition)
        matrix_metadata.update(self.user_metadata)

        return(matrix_metadata)

    def generate_plans(self, matrix_set_definitions, feature_dictionaries):
        """Create build tasks and update the matrix definitions with UUIDs

        :param matrix_set_definitions: the temporal information needed to generate each matrix
        :param feature_dictionaries: combinations of features to include in matrices
        :type matrix_set_definitions: list
        :type feature_dictionaries: list

        :return: matrix set definitions (updated with matrix uuids) and build tasks
        :rtype: tuple (list, dict)
        """
        updated_definitions = []
        build_tasks = dict()
        for matrix_set in matrix_set_definitions:
            logging.info('Making plans for matrix set %s', matrix_set)
            logging.info(
                'Iterating over %s label names, %s label_types, %s states, '
                '%s feature dictionaries',
                len(self.label_names),
                len(self.label_types),
                len(self.states),
                len(feature_dictionaries)
            )
            train_matrix = matrix_set['train_matrix']
            for (
                label_name,
                label_type,
                state,
                feature_dictionary,
            ) in itertools.product(
                self.label_names,
                self.label_types,
                self.states,
                feature_dictionaries
            ):
                matrix_set_clone = copy.deepcopy(matrix_set)
                # get a uuid
                train_metadata = self._make_metadata(
                    train_matrix,
                    feature_dictionary,
                    label_name,
                    label_type,
                    state,
                    'train',
                )
                train_uuid = metta.generate_uuid(train_metadata)
                logging.info('Matrix UUID %s found for train metadata %s',
                             train_uuid, train_metadata)
                if train_uuid not in build_tasks:
                    build_tasks[train_uuid] = self._generate_build_task(
                        train_metadata,
                        train_uuid,
                        train_matrix,
                        feature_dictionary
                    )
                    logging.info('Train uuid %s not found in build tasks yet, '
                                 'so added', train_uuid)
                else:
                    logging.info('Train uuid %s already found in build tasks', train_uuid)
                matrix_set_clone['train_uuid'] = train_uuid

                test_uuids = []
                for test_matrix in matrix_set_clone['test_matrices']:
                    test_metadata = self._make_metadata(
                        test_matrix,
                        feature_dictionary,
                        label_name,
                        label_type,
                        state,
                        'test',
                    )
                    test_uuid = metta.generate_uuid(test_metadata)
                    logging.info('Matrix UUID %s found for test metadata %s',
                                 test_uuid, test_metadata)
                    if test_uuid not in build_tasks:
                        build_tasks[test_uuid] = self._generate_build_task(
                            test_metadata,
                            test_uuid,
                            test_matrix,
                            feature_dictionary
                        )
                        logging.info('Test uuid %s not found in build tasks '
                                     'yet, so added', test_uuid)
                    else:
                        logging.info('Test uuid %s already found in build tasks',
                                     test_uuid)

                    test_uuids.append(test_uuid)
                matrix_set_clone['test_uuids'] = test_uuids
                updated_definitions.append(matrix_set_clone)

        logging.info(
            'Planner is finished generating matrix plans. '
            '%s matrix definitions and %s unique build tasks found',
            len(updated_definitions),
            len(build_tasks.keys())
        )
        return updated_definitions, build_tasks
